# Intelligent Vulnerability Report Aggregation System

This project is an advanced vulnerability management system that aggregates reports from multiple security scanners, deduplicates findings, and generates a consolidated view of vulnerabilities. The system uses natural language processing (NLP) and machine learning techniques to provide more meaningful vulnerability assessments with less redundancy.



![image-20250327223232012](./.img/image-20250327223232012.png)



## Features

- Parsing multiple vulnerability scanner formats (NMAP, Nuclei, OpenVAS)
- Semantic deduplication using NLP embeddings 
- Automatic correlation of findings across different scanners
- Consolidated JSON output with comprehensive vulnerability information
- Intelligent merging of duplicate vulnerabilities while preserving all relevant information

## System Architecture

The system employs a modular pipeline architecture consisting of three main components:

1. **Parsing Module**: Extracts and normalizes data from various scanner report formats
2. **Embedding Module**: Transforms vulnerability descriptions into semantic vector representations using the MiniLM model
3. **Deduplication Module**: Uses DBSCAN clustering algorithm to identify and merge similar vulnerabilities

## Prerequisites

- Python 3.8 or higher
- Pip package manager
- 1GB+ of free disk space (for model downloads)

## Installation

### 1. Clone the repository

```bash
git clone https://github.com/yourusername/vulnerability-aggregation-system.git
cd vulnerability-aggregation-system
```

### 2. Create and activate a virtual environment

```bash
# Create virtual environment
python3 -m venv venv

# Activate virtual environment
# On Linux/macOS:
source venv/bin/activate
# On Windows:
venv\Scripts\activate
```

### 3. Install dependencies

```bash
pip install -r requirements.txt
```

This will install the necessary packages including:
- pydantic (data validation)
- numpy and scipy (scientific computing)
- scikit-learn (for DBSCAN clustering)
- sentence-transformers (for embeddings)
- torch (required by sentence-transformers)
- transformers (required by sentence-transformers)
- aiohttp (for asynchronous operations)
- pytest (for testing)

### 4. Download the embedding model (optional)

The system will automatically download the required embedding model on first use, but you can pre-download it with:

```bash
python -m src.utils.model_downloader
```

## Usage

### 1. Prepare Input Data

Place your vulnerability scanner reports in the `data/input` directory. The system supports:
- NMAP XML output files (generated with `-oX` flag)
- Nuclei JSON/TXT reports
- OpenVAS XML reports

Example:
```bash
# Create directories if they don't exist
mkdir -p data/input data/output

# Copy your scanner reports to the input directory
cp path/to/your/nmap_report.xml data/input/
cp path/to/your/nuclei_report.txt data/input/
cp path/to/your/openvas_report.txt data/input/
```

### 2. Run the Program

```bash
python -m src.main
```

By default, the program will:
1. Process all files in the `data/input` directory
2. Parse each file with the appropriate scanner parser
3. Generate embeddings for vulnerability descriptions
4. Deduplicate similar vulnerabilities
5. Save a consolidated JSON report to the `data/output` directory

### 3. View Results

The consolidated report will be saved in the `data/output` directory with a timestamped filename (e.g., `vulnerabilities_20250327_120145.json`).

## Custom Configuration

You can customize the output directory:

```bash
python -m src.main --output-dir /path/to/custom/output
```

## Project Structure

- `src/`: Main source code
  - `parsers/`: Scanner-specific parsers (NMAP, Nuclei, OpenVAS)
  - `models/`: Data models for vulnerabilities
  - `analysis/`: Embedding and deduplication algorithms
  - `utils/`: Utility functions and helpers
- `data/`: Input and output data directories
- `config/`: Configuration files

## Example

### Input

**nmap:** 

![image-20250327220641216](./.img/image-20250327220641216.png)

**Nuclei:** 

![image-20250327220700462](./.img/image-20250327220700462.png)

**OpenVas:**

![image-20250327220826417](./.img/image-20250327220826417.png)



### Output

**JSON**:

![image-20250327220900652](./.img/image-20250327220900652.png)



## Author & Status

This project was developed by Soufiane EL KHARMOUDI as part of an academic project at HEPIA/HES-SO. It is currently in early development, with ongoing improvements focused on accuracy, scanner format support, and deduplication algorithms.

## License

This project is licensed under the MIT License - see the [LICENSE](LICENSE) file for details.