import re
from typing import List, Dict, Any

from ..models.vulnerability import Vulnerability, Severity, AffectedComponent
from ..utils.logging import get_logger


class VulnerabilityCorrelator:
    """
    Correlates findings from different vulnerability scanners to enhance detection.

    Different scanners often detect related aspects of the same underlying vulnerability.
    For example:
    - Nmap might detect a vulnerable service version
    - Nuclei might identify a specific CVE affecting that service
    - ZAP might find the exploitable endpoint

    This correlator connects these findings by:
    1. Grouping vulnerabilities affecting the same target (IP:port)
    2. Identifying service information from infrastructure scanners
    3. Matching CVE details from vulnerability scanners
    4. Cross-referencing findings to build a more complete picture
    5. Enhancing vulnerability descriptions with correlated information

    This provides a more comprehensive view of each vulnerability.
    """

    def __init__(self):
        self.logger = get_logger("VulnerabilityCorrelator")

    async def correlate_findings(
        self, vulnerabilities: List[Vulnerability]
    ) -> List[Vulnerability]:
        """
        Correlate findings from different scanners based on affected components.

        Args:
            vulnerabilities: List of vulnerabilities to correlate

        Returns:
            List of vulnerabilities with enhanced correlation
        """
        self.logger.info(f"Correlating {len(vulnerabilities)} vulnerabilities")

        # Group vulnerabilities by IP address and port
        location_groups = {}
        for vuln in vulnerabilities:
            for component in vuln.affected_components:
                if component.location:
                    # Try to extract IP and port
                    match = re.search(
                        r"(\d+\.\d+\.\d+\.\d+)(?::(\d+))?", component.location
                    )
                    if match:
                        ip = match.group(1)
                        port = match.group(2) or "unknown"
                        key = f"{ip}:{port}"

                        if key not in location_groups:
                            location_groups[key] = []
                        location_groups[key].append(vuln)

        # For each location, enrich vulnerabilities with correlated info
        for location, loc_vulns in location_groups.items():
            if len(loc_vulns) <= 1:
                continue

            # Find service information (usually from Nmap)
            service_info = None
            for vuln in loc_vulns:
                if "Nmap" in vuln.discovered_by and vuln.severity == Severity.INFO:
                    service_info = vuln
                    break

            # Find CVE information (usually from Nuclei)
            cve_vulns = []
            for vuln in loc_vulns:
                if vuln.id and vuln.id.startswith("CVE-") and vuln != service_info:
                    cve_vulns.append(vuln)

            # If we have both service info and CVE info, cross-reference them
            if service_info and cve_vulns:
                for cve_vuln in cve_vulns:
                    # Add service details to CVE vulnerability
                    for component in service_info.affected_components:
                        if component not in cve_vuln.affected_components:
                            cve_vuln.affected_components.append(component)

                    # Add context to description
                    if not cve_vuln.description.endswith("\n"):
                        cve_vuln.description += "\n"
                    cve_vuln.description += f"\n### Service Information\n"
                    cve_vuln.description += f"This vulnerability affects {' '.join(comp.name + ' ' + (comp.version or '') for comp in service_info.affected_components)}.\n"

        self.logger.info("Correlation completed")
        return vulnerabilities
