from pathlib import Path
from typing import List, Dict, Optional, Set
import numpy as np
from sklearn.cluster import DBSCAN
import re
from datetime import datetime
import logging

from ..models.vulnerability import CVSSMetrics, Vulnerability, Severity
from .embedding import VulnerabilityEmbedder


class VulnerabilityDeduplicator:
    """System for identifying and merging duplicate vulnerabilities using embeddings and clustering."""

    def __init__(
        self,
        similarity_threshold: float = 0.85,
        min_samples: int = 2,
        model_path: Optional[str] = None,
    ):
        """
        Initialize the deduplicator.

        Args:
            similarity_threshold: Threshold for similarity comparison (0.0-1.0)
            min_samples: Minimum samples required to form a cluster
            model_path: Optional path to a local embedding model
        """
        # Check for default model directory
        if model_path is None:
            default_model_dir = Path.cwd() / "models" / "sentence-transformer"
            if default_model_dir.exists():
                model_path = str(default_model_dir)

        self.embedder = VulnerabilityEmbedder(model_path=model_path)
        self.similarity_threshold = similarity_threshold
        self.min_samples = min_samples
        self.logger = logging.getLogger(__name__)

        # Configure detailed logging
        self.logger.info(
            f"Initialized VulnerabilityDeduplicator with "
            f"similarity_threshold={similarity_threshold}, "
            f"min_samples={min_samples}"
        )

    async def deduplicate(
        self, vulnerabilities: List[Vulnerability]
    ) -> List[Vulnerability]:
        """
        Identify and merge duplicate vulnerabilities.

        Args:
            vulnerabilities: List of vulnerabilities to deduplicate

        Returns:
            List of unique vulnerabilities with duplicates merged
        """
        if not vulnerabilities:
            return []

        # Check if the dataset is very large and should use batched processing
        if len(vulnerabilities) > 2000:  # Arbitrary threshold, adjust as needed
            self.logger.info(
                f"Large dataset detected ({len(vulnerabilities)} vulnerabilities). Using batched processing."
            )
            return await self.deduplicate_large(vulnerabilities)

        # Standard processing for smaller datasets
        # Generate embeddings
        embeddings_dict = self.embedder.generate_embeddings(vulnerabilities)

        # Convert to numpy array maintaining order
        vuln_ids = [v.id for v in vulnerabilities if v.id in embeddings_dict]
        if not vuln_ids:
            return vulnerabilities  # Return original list if no embeddings generated

        embeddings_array = np.array([embeddings_dict[vid] for vid in vuln_ids]).reshape(
            len(vuln_ids), -1
        )  # Reshape to 2D array

        # Perform clustering
        clusters = self._cluster_vulnerabilities(embeddings_array)

        # Group vulnerabilities by cluster
        cluster_groups = self._group_by_cluster(vulnerabilities, vuln_ids, clusters)

        # Merge vulnerabilities in each cluster
        unique_vulns = []
        for group in cluster_groups:
            if len(group) == 1:
                unique_vulns.append(group[0])
            else:
                merged = self._merge_vulnerabilities(group)
                unique_vulns.append(merged)

        return unique_vulns

    async def deduplicate_large(
        self, vulnerabilities: List[Vulnerability], batch_size: int = 1000
    ) -> List[Vulnerability]:
        """
        Deduplicate a large set of vulnerabilities using batching.

        For very large datasets, this method splits the processing into
        batches to avoid memory issues.

        Args:
            vulnerabilities: List of vulnerabilities to deduplicate
            batch_size: Size of each processing batch

        Returns:
            List of unique vulnerabilities with duplicates merged
        """
        if len(vulnerabilities) <= batch_size:
            # If dataset is small enough, use standard deduplication
            return await self.deduplicate(vulnerabilities)

        self.logger.info(
            f"Starting batched deduplication of {len(vulnerabilities)} vulnerabilities "
            f"using batch size of {batch_size}"
        )

        # Split into batches
        batches = [
            vulnerabilities[i : i + batch_size]
            for i in range(0, len(vulnerabilities), batch_size)
        ]

        # Process first batch
        result = await self.deduplicate(batches[0])
        self.logger.info(
            f"Processed batch 1/{len(batches)} - {len(result)} unique vulnerabilities"
        )

        # Process remaining batches by comparing to results so far
        for i, batch in enumerate(batches[1:], 2):
            # Generate embeddings for this batch
            batch_embeddings = self.embedder.generate_embeddings(batch)

            # Generate embeddings for current results
            result_embeddings = self.embedder.generate_embeddings(result)

            # Identify matches between this batch and existing results
            matched_ids = set()
            for batch_id, batch_vector in batch_embeddings.items():
                for result_id, result_vector in result_embeddings.items():
                    similarity = self.embedder.calculate_similarity(
                        batch_vector, result_vector
                    )
                    if similarity >= self.similarity_threshold:
                        matched_ids.add(batch_id)
                        # Find the vulnerability objects
                        batch_vuln = next(v for v in batch if v.id == batch_id)
                        result_vuln = next(v for v in result if v.id == result_id)
                        # Merge batch vuln into result vuln
                        merged = self._merge_vulnerabilities([result_vuln, batch_vuln])
                        # Replace the original in results
                        result = [v for v in result if v.id != result_id]
                        result.append(merged)
                        break

            # Add unmatched vulnerabilities from this batch to the results
            unmatched = [v for v in batch if v.id not in matched_ids]
            if unmatched:
                # Deduplicate within unmatched items
                unmatched_unique = await self.deduplicate(unmatched)
                result.extend(unmatched_unique)

            self.logger.info(
                f"Processed batch {i}/{len(batches)} - Now have {len(result)} unique vulnerabilities"
            )

        return result

    def _cluster_vulnerabilities(self, embeddings: np.ndarray) -> np.ndarray:
        """
        Cluster vulnerabilities based on their embeddings using DBSCAN algorithm.

        DBSCAN (Density-Based Spatial Clustering of Applications with Noise) is used here
        because:
        1. It doesn't require specifying the number of clusters in advance
        2. It can find arbitrarily shaped clusters
        3. It can identify outliers as noise points

        The 'eps' parameter represents the maximum distance between two samples to be
        considered in the same neighborhood. It's derived from our similarity threshold
        (1 - similarity_threshold) since we're using cosine distance.

        The 'min_samples' parameter is the minimum number of samples in a neighborhood
        for a point to be considered a core point (including the point itself).

        Args:
            embeddings: Array of vulnerability embeddings

        Returns:
            Array of cluster labels
        """
        # Calculate distance threshold from similarity threshold
        distance_threshold = 1 - self.similarity_threshold

        # Perform DBSCAN clustering
        clustering = DBSCAN(
            eps=distance_threshold, min_samples=self.min_samples, metric="cosine"
        )

        return clustering.fit_predict(embeddings)

    def _group_by_cluster(
        self,
        vulnerabilities: List[Vulnerability],
        vuln_ids: List[str],
        clusters: np.ndarray,
    ) -> List[List[Vulnerability]]:
        """
        Group vulnerabilities by their cluster assignments.

        Args:
            vulnerabilities: List of vulnerabilities
            vuln_ids: List of vulnerability IDs in order
            clusters: Array of cluster labels

        Returns:
            List of vulnerability groups
        """
        # Create mapping of vulnerability ID to vulnerability
        vuln_map = {v.id: v for v in vulnerabilities if v.id}

        # Group by cluster
        groups = {}
        for vid, cluster_id in zip(vuln_ids, clusters):
            if cluster_id not in groups:
                groups[cluster_id] = []
            groups[cluster_id].append(vuln_map[vid])

        return list(groups.values())

    def _merge_vulnerabilities(self, vulns: List[Vulnerability]) -> Vulnerability:
        """Merge a group of similar vulnerabilities into one."""
        # Use vulnerability with CVE as base, or most severe if no CVE
        base_with_cve = next(
            (v for v in vulns if v.id and v.id.startswith("CVE-")), None
        )

        if base_with_cve:
            base = base_with_cve
        else:
            # Otherwise, use vulnerability with highest severity
            base = max(vulns, key=lambda v: v.severity)

        # Collect unique information
        cves = set()
        affected_components = set()
        references = set()
        discovered_by = set()
        tags = set()

        # Extract potential CVEs from descriptions
        cve_pattern = r"CVE-\d{4}-\d+"

        for vuln in vulns:
            # Extract CVEs from descriptions
            found_cves = re.findall(cve_pattern, vuln.description)
            cves.update(found_cves)

            # Collect other information
            affected_components.update(vuln.affected_components)
            references.update(vuln.references)
            discovered_by.update(vuln.discovered_by)
            tags.update(vuln.tags)

        # Create consolidated description
        consolidated_description = self._create_structured_description(
            vulns, cves, affected_components
        )

        # Get best CVSS data from all vulnerabilities
        best_cvss = self._get_best_cvss(vulns)

        # Create the merged vulnerability
        result = Vulnerability(
            id=base.id,
            title=f"Security Issue: {base.title}",
            description=consolidated_description,
            severity=max(v.severity for v in vulns),
            cvss=best_cvss,
            affected_components=sorted(
                affected_components, key=lambda x: (x.name, x.version or "")
            ),
            discovered_by=sorted(discovered_by),
            discovery_date=min(v.discovery_date for v in vulns),
            last_seen_date=max(v.last_seen_date for v in vulns),
            references=sorted(references),
            tags=sorted(tags),
            metadata={
                "merged_from": [v.id for v in vulns if v.id],
                "merge_date": datetime.now().isoformat(),
                "related_cves": sorted(cves),
                "source_count": len(discovered_by),
                "merge_confidence": self._calculate_merge_confidence(vulns),
            },
        )

        # Ensure that composite_id is preserved or recalculated
        if any(v.composite_id for v in vulns):
            # Use the most specific composite ID (the one with fewest "unknown")
            composite_ids = [v.composite_id for v in vulns if v.composite_id]
            if composite_ids:
                best_composite_id = min(
                    composite_ids,
                    key=lambda cid: cid.count("unknown") if cid else float("inf"),
                )
                result.composite_id = best_composite_id

        return result

    def _calculate_merge_confidence(self, vulns: List[Vulnerability]) -> float:
        """
        Calculate confidence score for the merge operation.

        Args:
            vulns: List of merged vulnerabilities

        Returns:
            Confidence score between 0 and 1
        """
        factors = []

        # Check if vulnerabilities share CVEs
        has_shared_cves = (
            len({v.id for v in vulns if v.id and v.id.startswith("CVE-")}) > 1
        )
        if has_shared_cves:
            factors.append(1.0)

        # Check if vulnerabilities affect same components
        component_names = {c.name for v in vulns for c in v.affected_components}
        if len(component_names) == 1:
            factors.append(1.0)
        elif len(component_names) > 1:
            factors.append(0.5)

        # Check severity consistency
        severity_values = {v.severity for v in vulns}
        if len(severity_values) == 1:
            factors.append(1.0)
        else:
            factors.append(0.7)

        # Calculate final confidence
        return sum(factors) / len(factors) if factors else 0.5

    def _create_structured_description(self, vulns, cves, affected_components):
        """Create a well-organized description with logical sections."""
        # Start with a summary
        description = "## Vulnerability Summary\n\n"

        # Add CVE information
        if cves:
            description += f"Related CVEs: {', '.join(sorted(cves))}\n\n"

        # Add affected services
        description += "## Affected Services\n"
        for component in sorted(affected_components, key=lambda x: x.name):
            component_info = f"- {component.name}"
            if component.version:
                component_info += f" {component.version}"
            if component.location:
                component_info += f" at {component.location}"
            description += f"{component_info}\n"

        # Organize findings by source to avoid duplication
        findings_by_source = {}
        for vuln in vulns:
            for source in vuln.discovered_by:
                if source not in findings_by_source:
                    findings_by_source[source] = set()

                # Extract key information and avoid duplication
                key_info = self._extract_key_information(vuln.description)
                if key_info:
                    findings_by_source[source].add(key_info)

        # Add findings from each source
        description += "\n## Detailed Findings\n"
        for source, findings in findings_by_source.items():
            for finding in sorted(findings):
                description += f"\nSource: {source}\n{finding}\n"

        # Add recommendations if available
        remediation_info = self._extract_remediation_info(vulns)
        if remediation_info:
            description += "\n## Recommendations\n"
            description += remediation_info

        return description

    def _extract_key_information(self, description):
        """Extract key technical information from a description."""
        if not description:
            return ""

        lines = description.split("\n")
        result = []
        skip_line = False

        for line in lines:
            # Skip redundant headers
            if re.match(r"^#+\s+(Summary|Description|Details):", line, re.IGNORECASE):
                skip_line = False
                continue

            # Skip metadata lines
            if re.match(r"^(Severity|Risk|CVE ID|Confidence):", line, re.IGNORECASE):
                continue

            result.append(line)

        return "\n".join(result)

    def _extract_remediation_info(self, vulns):
        """Extract remediation information from vulnerabilities."""
        remediation = []

        for vuln in vulns:
            # Look for remediation sections
            sections = re.findall(
                r"(?:Remediation|Solution|Recommendation)s?:?\s*(.*?)(?:\n\n|\Z)",
                vuln.description,
                re.IGNORECASE | re.DOTALL,
            )
            for section in sections:
                if section and section.strip() and section.strip() not in remediation:
                    remediation.append(section.strip())

        return "\n\n".join(remediation)

    def _get_best_cvss(self, vulns: List[Vulnerability]) -> Optional[CVSSMetrics]:
        """
        Select the best CVSS metrics from a list of vulnerabilities.

        Args:
            vulns: List of vulnerabilities

        Returns:
            The most relevant/severe CVSS metrics or None if no CVSS metrics are available
        """
        # Filter vulnerabilities that have CVSS metrics
        vulns_with_cvss = [v for v in vulns if v.cvss is not None]

        if not vulns_with_cvss:
            return None

        # Define version preference (higher is better)
        def get_version_weight(version: str) -> float:
            if version.startswith("3.1"):
                return 3.1
            elif version.startswith("3.0"):
                return 3.0
            elif version.startswith("2.0"):
                return 2.0
            else:
                # Try to extract a numeric version
                try:
                    return float(version.split()[0])
                except:
                    return 0.0

        # Sort by version (descending) then base score (descending)
        sorted_vulns = sorted(
            vulns_with_cvss,
            key=lambda v: (get_version_weight(v.cvss.version), v.cvss.base_score),
            reverse=True,
        )

        # Return CVSS from first vulnerability (highest version and score)
        return sorted_vulns[0].cvss
