from datetime import time
from typing import List, Dict, Optional
import numpy as np
import logging
import os
import ssl
import psutil
from pathlib import Path
from sentence_transformers import SentenceTransformer
from sklearn.decomposition import TruncatedSVD
from ..models.vulnerability import Vulnerability


class VulnerabilityEmbedder:
    """Generate embeddings for vulnerabilities using transformer models."""

    def __init__(
        self,
        model_name: str = "all-MiniLM-L6-v2",
        model_path: Optional[str] = None,
        use_fallback: bool = True,
    ):
        """
        Initialize the embedding model using a multi-stage loading strategy.

        The embedding model is used to convert vulnerability text into dense vector
        representations for similarity comparison. The system tries multiple methods
        to load the model in this priority order:

        1. From a local path if specified (fastest, works offline)
        2. From Hugging Face using standard loading (requires internet)
        3. From Hugging Face with SSL verification disabled (for problematic networks)
        4. Fallback to a simple TF-IDF + SVD approach (much lower quality but works anywhere)

        Args:
            model_name: Name of the model to use if downloading from Hugging Face.
                    Default is "all-MiniLM-L6-v2" which balances size and quality.
            model_path: Path to a locally stored model. This is preferred if available.
            use_fallback: Whether to use fallback methods if model loading fails.
                        Setting this to False will raise an error if the model cannot be loaded.
        """
        self.logger = logging.getLogger(__name__)
        self.model = None
        self.embedding_dim = 384  # Default for MiniLM-L6-v2
        self.using_fallback = False

        # Try different methods to load the model
        if model_path and os.path.exists(model_path):
            # Try loading from specified local path
            self._load_from_local_path(model_path)
        else:
            # Try standard loading
            self._try_standard_loading(model_name)

            # If still no model, try with SSL verification disabled
            if self.model is None and use_fallback:
                self._try_ssl_workaround(model_name)

            # If still no model, use fallback method
            if self.model is None and use_fallback:
                self._setup_fallback()

    def _load_from_local_path(self, model_path: str) -> None:
        """Try to load the model from a local path."""
        try:
            self.logger.info(f"Attempting to load model from local path: {model_path}")
            self.model = SentenceTransformer(model_path)
            self.logger.info("Successfully loaded model from local path")
        except Exception as e:
            self.logger.error(f"Failed to load model from local path: {e}")

    def _try_standard_loading(self, model_name: str) -> None:
        """Try to load the model using standard methods."""
        try:
            self.logger.info(f"Attempting to load model: {model_name}")
            self.model = SentenceTransformer(model_name)
            self.logger.info("Successfully loaded model")
        except Exception as e:
            self.logger.error(f"Failed to load model using standard method: {e}")

    def _try_ssl_workaround(self, model_name: str) -> None:
        """Try to load the model with SSL verification disabled."""
        try:
            self.logger.info("Attempting to load model with SSL verification disabled")
            # Temporarily disable SSL verification
            old_https_context = ssl._create_default_https_context
            ssl._create_default_https_context = ssl._create_unverified_context

            try:
                self.model = SentenceTransformer(model_name)
                self.logger.info(
                    "Successfully loaded model with SSL verification disabled"
                )
            finally:
                # Restore original SSL context
                ssl._create_default_https_context = old_https_context
        except Exception as e:
            self.logger.error(f"Failed to load model even with SSL workaround: {e}")

    def _setup_fallback(self) -> None:
        """
        Set up the fallback embedding method when transformer models cannot be loaded.

        This fallback uses a classical NLP approach combining TF-IDF vectorization
        (Term Frequency-Inverse Document Frequency) with SVD (Singular Value Decomposition)
        dimensionality reduction. While significantly lower quality than transformer embeddings,
        this approach:

        1. Works without external dependencies beyond sklearn
        2. Runs on any hardware (no GPU or large memory needed)
        3. Is deterministic and doesn't require internet access
        4. Provides a reasonable baseline for text similarity

        The main drawbacks are:
        1. Much less semantic understanding
        2. More affected by superficial text differences
        3. Lower accuracy for vulnerability deduplication
        """
        self.logger.warning(
            "Using simple fallback embedding method (significantly lower quality)"
        )
        self.using_fallback = True

        # Here we'll implement a simple TF-IDF + SVD approach
        try:
            from sklearn.feature_extraction.text import TfidfVectorizer
            from sklearn.decomposition import TruncatedSVD

            self.vectorizer = TfidfVectorizer(max_features=5000, stop_words="english")
            self.svd = TruncatedSVD(n_components=self.embedding_dim)

            # The actual fitting will happen when the first batch comes in
            self.logger.info("Fallback embedding method initialized")
        except ImportError:
            self.logger.error(
                "sklearn not available for fallback method. Please install it with 'pip install scikit-learn'"
            )
            raise ImportError("sklearn required for fallback embedding method")

    def generate_embeddings(
        self, vulnerabilities: List[Vulnerability]
    ) -> Dict[str, np.ndarray]:
        """Generate embeddings for multiple vulnerabilities."""
        if not vulnerabilities:
            return {}

        self.logger.info(
            f"Generating embeddings for {len(vulnerabilities)} vulnerabilities"
        )

        # Prepare texts for batch encoding
        texts = []
        ids = []
        skipped = 0

        for vuln in vulnerabilities:
            if not vuln.id:
                skipped += 1
                continue

            self.logger.debug(f"Processing vulnerability: {vuln.id}")
            text = f"{vuln.title} {vuln.description}"

            if vuln.affected_components:
                components = " ".join(
                    f"{c.name} {c.version or ''}" for c in vuln.affected_components
                )
                text += f" {components}"

            texts.append(text)
            ids.append(vuln.id)

        if skipped > 0:
            self.logger.warning(f"Skipped {skipped} vulnerabilities without ID")

        if not texts:
            self.logger.warning("No valid texts to encode")
            return {}

        # Generate embeddings
        if self.model is not None:
            # Use the transformer model
            embeddings = self.model.encode(texts)
        elif self.using_fallback:
            # Use the fallback method
            embeddings = self._generate_embeddings_fallback(texts)
        else:
            self.logger.error("No embedding method available")
            return {}

        self.logger.debug(f"Generated embeddings shape: {embeddings.shape}")

        # Create mapping
        result = {vuln_id: embedding for vuln_id, embedding in zip(ids, embeddings)}

        self.logger.info(f"Successfully generated {len(result)} embeddings")
        return result

    def _generate_embeddings_fallback(self, texts: List[str]) -> np.ndarray:
        """Generate embeddings using the fallback method."""
        # Fit the vectorizer on this batch
        tfidf_matrix = self.vectorizer.fit_transform(texts)

        # Determine max possible dimensions based on the data
        n_components = min(384, tfidf_matrix.shape[1], tfidf_matrix.shape[0] - 1)

        if n_components < 2:
            # Not enough features for SVD, just return normalized TF-IDF vectors
            embeddings = tfidf_matrix.toarray()
        else:
            # Adjust SVD to use appropriate dimensions
            self.svd = TruncatedSVD(n_components=n_components)
            embeddings = self.svd.fit_transform(tfidf_matrix)

        # Normalize embeddings (cosine similarity needs unit vectors)
        norms = np.linalg.norm(embeddings, axis=1, keepdims=True)
        norms[norms == 0] = 1  # Avoid division by zero
        normalized_embeddings = embeddings / norms

        return normalized_embeddings

    def calculate_similarity(
        self, embedding1: np.ndarray, embedding2: np.ndarray
    ) -> float:
        """
        Calculate cosine similarity between two embeddings.

        Args:
            embedding1: First embedding
            embedding2: Second embedding

        Returns:
            Similarity score between 0 and 1
        """
        return np.dot(embedding1, embedding2) / (
            np.linalg.norm(embedding1) * np.linalg.norm(embedding2)
        )

    def save_model(self, path: str) -> bool:
        """
        Save the currently loaded model to a local directory.

        Args:
            path: Directory path to save the model

        Returns:
            True if successful, False otherwise
        """
        if self.model is None:
            self.logger.error("No model to save")
            return False

        try:
            os.makedirs(path, exist_ok=True)
            self.model.save(path)
            self.logger.info(f"Model saved to {path}")
            return True
        except Exception as e:
            self.logger.error(f"Error saving model: {e}")
            return False
