"""
Module for classifying vulnerabilities using both keyword matching and embeddings.
"""

import logging
from typing import Dict, List, Optional, Tuple
import numpy as np
from pathlib import Path
import json
from datetime import datetime

from ..models.vulnerability import Vulnerability, Severity
from .embedding import VulnerabilityEmbedder


class VulnerabilityClassifier:
    """Advanced classifier for vulnerabilities using both keywords and embeddings."""

    _instance = None

    @classmethod
    def get_instance(cls, similarity_threshold: float = 0.75):
        """Get or create the singleton instance of the classifier."""
        if cls._instance is None:
            cls._instance = cls(similarity_threshold)
        return cls._instance

    def __init__(self, similarity_threshold: float = 0.75):
        """
        Initialize the vulnerability classifier.

        Args:
            similarity_threshold: Threshold for AI classification confidence
        """
        self.logger = logging.getLogger(__name__)
        self.embedder = VulnerabilityEmbedder()
        self.similarity_threshold = similarity_threshold

        self.vuln_types = self._load_vulnerability_types()

        # Cache
        self._reference_embeddings = {}

    def _load_vulnerability_types(self) -> Dict[str, List[str]]:
        """Load vulnerability types from a JSON file or use the default dictionary."""
        try:
            # test!!!
            json_path = (
                Path(__file__).parent.parent / "data" / "vulnerability_types.json"
            )

            if json_path.exists():
                self.logger.info(f"Loading vulnerability types from {json_path}")
                with open(json_path, "r") as f:
                    return json.load(f)
            else:
                self.logger.warning(f"JSON file not found at {json_path}")
        except Exception as e:
            self.logger.warning(f"Could not load vulnerability types from file: {e}")

        # If file loading failed, use a minimal built-in dictionary
        self.logger.info("Using built-in vulnerability types dictionary")
        return {
            "sqli": ["sql injection", "sqli", "sql", "database injection"],
            "xss": [
                "cross-site scripting",
                "xss",
                "client-side script",
                "javascript injection",
            ],
            "rce": ["remote code execution", "rce", "code execution", "arbitrary code"],
            "csrf": ["cross-site request forgery", "csrf", "xsrf", "sea surf"],
            "ssrf": ["server-side request forgery", "ssrf", "request forgery"],
            "lfi": ["local file inclusion", "lfi", "file inclusion", "path inclusion"],
            "path_traversal": [
                "path traversal",
                "directory traversal",
                "path manipulation",
            ],
            "auth_bypass": ["authentication bypass", "auth bypass", "login bypass"],
            "idor": [
                "insecure direct object reference",
                "idor",
                "direct object reference",
            ],
            "info_disclosure": [
                "information disclosure",
                "information leakage",
                "data exposure",
            ],
            "weak_crypto": ["weak cryptography", "weak encryption", "weak cipher"],
            "misconfiguration": [
                "misconfiguration",
                "default configuration",
                "insecure configuration",
            ],
            "dos": ["denial of service", "dos", "resource exhaustion"],
            "outdated_software": [
                "outdated",
                "obsolete",
                "old version",
                "vulnerable version",
            ],
            "weak_password": ["weak password", "password policy", "default password"],
        }

    def classify(self, vulnerability: Vulnerability) -> Tuple[str, float, str]:
        """
        Classify a vulnerability using both keywords and AI.

        Args:
            vulnerability: The vulnerability to classify

        Returns:
            Tuple of (vulnerability_type, confidence_score, method_used)
        """
        # Step 1: Try keyword matching first
        vuln_type, confidence, method = self._classify_by_keywords(vulnerability)

        # If keyword matching found something with high confidence, return it
        if confidence > 0.9:
            return vuln_type, confidence, method

        # Step 2: Try AI classification
        ai_type, ai_confidence = self._classify_by_ai(vulnerability)

        # Return the best classification
        if ai_confidence > confidence:
            return ai_type, ai_confidence, "ai"
        else:
            return vuln_type, confidence, method

    def _classify_by_keywords(
        self, vulnerability: Vulnerability
    ) -> Tuple[str, float, str]:
        """Classify a vulnerability using keyword matching."""
        text = (vulnerability.title + " " + vulnerability.description).lower()

        # Find all matching patterns
        matched_types = []
        for vuln_type, patterns in self.vuln_types.items():
            for pattern in patterns:
                if pattern in text:
                    # Calculate a confidence score based on pattern length and specificity
                    confidence = min(1.0, 0.7 + (len(pattern) / 50))
                    matched_types.append((vuln_type, pattern, confidence))

        if matched_types:
            # Sort by confidence and pattern length
            matched_types.sort(key=lambda x: (x[2], len(x[1])), reverse=True)
            return matched_types[0][0], matched_types[0][2], "keyword"

        # No matches found
        return vulnerability.severity.name.lower(), 0.5, "fallback"

    def _classify_by_ai(self, vulnerability: Vulnerability) -> Tuple[str, float]:
        """Classify a vulnerability using AI embeddings."""
        try:
            # Verifica se l'embedder sta usando il metodo di fallback
            if (
                hasattr(self.embedder, "using_fallback")
                and self.embedder.using_fallback
            ):
                # Se stiamo usando il metodo di fallback, non tentare la classificazione AI
                self.logger.warning(
                    "Skipping AI classification because using fallback embedding method"
                )
                return "unknown", 0.0

            # Generate an embedding for the current vulnerability
            current_embedding = self._get_embedding_for_vuln(vulnerability)

            if current_embedding is None:
                return "unknown", 0.0

            # Get or generate embeddings for vulnerability types
            best_type = None
            best_similarity = -1

            for vuln_type in self.vuln_types.keys():
                type_embedding = self._get_reference_embedding(vuln_type)
                if type_embedding is not None:
                    similarity = self._calculate_similarity(
                        current_embedding, type_embedding
                    )
                    if similarity > best_similarity:
                        best_similarity = similarity
                        best_type = vuln_type

            if best_type and best_similarity > self.similarity_threshold:
                return best_type, best_similarity

            return "unknown", best_similarity

        except Exception as e:
            self.logger.error(f"AI classification failed: {e}")
            return "unknown", 0.0

    def _get_embedding_for_vuln(
        self, vulnerability: Vulnerability
    ) -> Optional[np.ndarray]:
        """Generate an embedding for a vulnerability."""
        # Create a copy with a temporary ID
        temp_vuln = Vulnerability(
            id="temp_current",
            title=vulnerability.title,
            description=vulnerability.description,
            severity=vulnerability.severity,
            discovery_date=vulnerability.discovery_date,
            last_seen_date=vulnerability.last_seen_date,
        )

        # Generate embedding
        embeddings = self.embedder.generate_embeddings([temp_vuln])

        if "temp_current" in embeddings:
            return embeddings["temp_current"]
        return None

    def _get_reference_embedding(self, vuln_type: str) -> Optional[np.ndarray]:
        """Get or generate a reference embedding for a vulnerability type."""
        if vuln_type in self._reference_embeddings:
            return self._reference_embeddings[vuln_type]

        if vuln_type not in self.vuln_types:
            return None

        # Create a prototype vulnerability for this type
        patterns = self.vuln_types[vuln_type][:5]  # Use up to 5 patterns
        pattern_desc = " ".join(patterns)

        prototype = Vulnerability(
            id=f"ref_{vuln_type}",
            title=f"{vuln_type.replace('_', ' ')} vulnerability",
            description=f"This is a {vuln_type.replace('_', ' ')} vulnerability. {pattern_desc}",
            severity=Severity.MEDIUM,  # Default severity
            discovery_date=datetime.now(),
            last_seen_date=datetime.now(),
        )

        # Generate embedding
        embeddings = self.embedder.generate_embeddings([prototype])

        if f"ref_{vuln_type}" in embeddings:
            # Cache the embedding for future use
            self._reference_embeddings[vuln_type] = embeddings[f"ref_{vuln_type}"]
            return self._reference_embeddings[vuln_type]

        return None

    def _calculate_similarity(
        self, embedding1: np.ndarray, embedding2: np.ndarray
    ) -> float:
        """Calculate cosine similarity between two embeddings."""
        return np.dot(embedding1, embedding2) / (
            np.linalg.norm(embedding1) * np.linalg.norm(embedding2)
        )
