#!/usr/bin/env python3
"""
VulnUnify - Main Entry Point

This module serves as the entry point for the VulnUnify application, handling command line
arguments, configuration loading, and orchestrating the vulnerability report processing flow.
VulnUnify aggregates vulnerability reports from multiple scanners, deduplicates findings,
and generates a consolidated JSON report.

Usage:
    python -m src.main
"""
import asyncio
from pathlib import Path
import json
import argparse

from .processor import process_vulnerability_reports


def parse_args():
    """
    Parse command line arguments.
    """
    parser = argparse.ArgumentParser(
        description="Process vulnerability scanner reports and generate consolidated JSON report."
    )
    parser.add_argument(
        "--output-dir",
        "-o",
        type=str,
        default="data/output",
        help="Directory for output JSON report",
    )
    return parser.parse_args()


def load_config() -> dict:
    """
    Load configuration from default config file.
    """
    config_path = Path("config/default_config.json")
    if not config_path.exists():
        print(f"Warning: Default config file not found at {config_path}")
        return {}

    with open(config_path, "r") as f:
        return json.load(f)


class VulnProcessor:
    """
    Process vulnerability reports and generate JSON report directly.
    """

    @staticmethod
    async def process_and_generate(output_dir: Path, config: dict) -> Path:
        """
        Process reports and generate JSON.
        """
        # Always use input directory
        input_dir = Path("data/input")

        # Process vulnerability reports
        json_result = await process_vulnerability_reports(input_dir, output_dir, config)

        if not json_result:
            raise ValueError("No vulnerability data generated")

        return json_result


if __name__ == "__main__":
    # Parse command line arguments
    args = parse_args()

    # Set output directory
    output_dir = Path(args.output_dir)

    # Load default configuration
    config = load_config()

    # Process reports
    try:
        processor = VulnProcessor()
        json_file = asyncio.run(processor.process_and_generate(output_dir, config))
        print(f"JSON report generated: {json_file}")
    except Exception as e:
        print(f"Error processing vulnerability reports: {e}")
