from enum import Enum
from typing import List, Optional, Dict, Any
from datetime import datetime
from pydantic import BaseModel
import re
import hashlib


class Severity(str, Enum):
    CRITICAL = "CRITICAL"
    HIGH = "HIGH"
    MEDIUM = "MEDIUM"
    LOW = "LOW"
    INFO = "INFO"


class CVSSMetrics(BaseModel):
    version: str
    base_score: float
    temporal_score: Optional[float] = None
    environmental_score: Optional[float] = None
    vector_string: str


class AffectedComponent(BaseModel):
    name: str
    version: Optional[str] = None
    location: Optional[str] = None

    def __hash__(self):
        """Make the class hashable for set operations"""
        return hash((self.name, self.version, self.location))

    def __eq__(self, other):
        """Define equality for hash operations"""
        if not isinstance(other, AffectedComponent):
            return False
        return (
            self.name == other.name
            and self.version == other.version
            and self.location == other.location
        )


class Vulnerability(BaseModel):
    id: Optional[str] = None  # CVE if available
    composite_id: Optional[str] = None  # Format: "host:port:vuln_type:hash"
    title: str
    description: str
    severity: Severity
    cvss: Optional[CVSSMetrics] = None
    affected_components: List[AffectedComponent] = []
    discovered_by: List[str] = []
    discovery_date: datetime
    last_seen_date: datetime
    references: List[str] = []
    tags: List[str] = []
    metadata: Dict[str, Any] = {}

    class Config:
        arbitrary_types_allowed = True

    def __init__(self, **data):
        super().__init__(**data)
        # Generate composite ID automatically if it doesn't exist already
        if not self.composite_id:
            self._generate_composite_id()

    def _generate_composite_id(self):
        """Generate a composite identifier based on host, port, and vulnerability type."""
        if self.affected_components:
            host = "unknown"
            port = "unknown"
            vuln_type = self._extract_vuln_type()

            # Extract host and port from the first affected component
            for component in self.affected_components:
                if component.location:
                    # Try to extract host:port
                    match = re.match(
                        r"^(?:https?://)?([^:/]+)(?::(\d+))?", component.location
                    )
                    if match:
                        host = match.group(1) or "unknown"
                        port = (
                            match.group(2) or "80"
                            if component.location.startswith("http:")
                            else "443"
                        )
                        break

            # Generate a hash of the description to differentiate vulnerabilities
            # of the same type on the same host:port
            desc_hash = hashlib.md5(self.description.encode()).hexdigest()[:8]

            self.composite_id = f"{host}:{port}:{vuln_type}:{desc_hash}"

            # If we don't have a CVE ID, create a human-readable ID based on the type
            if not self.id or not self.id.startswith("CVE-"):
                self.id = f"VLN-{vuln_type.upper()}-{desc_hash}"

    def _extract_vuln_type(self) -> str:
        """
        Extract vulnerability type from title and description using simple heuristics.
        """
        text = (self.title + " " + self.description).lower()

        # Very basic type detection with a few common types
        if "sql injection" in text or "sqli" in text:
            return "sqli"
        elif "cross-site scripting" in text or "xss" in text:
            return "xss"
        elif "remote code execution" in text or "rce" in text:
            return "rce"
        elif "path traversal" in text or "directory traversal" in text:
            return "path_traversal"

        # Default to severity
        return self.severity.name.lower()
