from abc import ABC, abstractmethod
from typing import List, Dict, Any, Optional
from pathlib import Path
import logging
import xml.etree.ElementTree as ET
import json
import re
from datetime import datetime

from ..models.vulnerability import Vulnerability, Severity


class BaseParser(ABC):
    """Base abstract class for vulnerability scanner parsers with shared utilities."""

    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)

    @abstractmethod
    async def parse_file(self, file_path: Path) -> List[Vulnerability]:
        """
        Parse a vulnerability scanner report file.

        Args:
            file_path: Path to the report file

        Returns:
            List of parsed Vulnerability objects
        """
        pass

    @abstractmethod
    async def validate_file(self, file_path: Path) -> bool:
        """
        Validate if the file is in the correct format for this parser.

        Args:
            file_path: Path to the file to validate

        Returns:
            True if file is valid, False otherwise
        """
        pass

    def normalize_severity(self, raw_severity: str) -> Severity:
        """
        Normalize severity string to standard Severity enum.

        Args:
            raw_severity: Raw severity string from scanner

        Returns:
            Normalized Severity enum
        """
        severity_mapping = {
            # Critical mappings
            "critical": Severity.CRITICAL,
            "crit": Severity.CRITICAL,
            "4": Severity.CRITICAL,
            # High mappings
            "high": Severity.HIGH,
            "hi": Severity.HIGH,
            "3": Severity.HIGH,
            # Medium mappings
            "medium": Severity.MEDIUM,
            "med": Severity.MEDIUM,
            "2": Severity.MEDIUM,
            # Low mappings
            "low": Severity.LOW,
            "1": Severity.LOW,
            # Info mappings
            "info": Severity.INFO,
            "information": Severity.INFO,
            "0": Severity.INFO,
        }

        normalized = severity_mapping.get(
            raw_severity.lower(), Severity.MEDIUM  # Default to MEDIUM if unknown
        )

        self.logger.debug(f"Normalized severity from {raw_severity} to {normalized}")
        return normalized

    def get_current_timestamp(self) -> datetime:
        """Get current timestamp in UTC"""
        return datetime.utcnow()

    # --- Utility methods ---

    async def safe_parse_xml(self, file_path: Path) -> Optional[ET.Element]:
        """
        Safely parse an XML file and return the root element.

        Args:
            file_path: Path to the XML file

        Returns:
            Root XML element or None if parsing fails
        """
        try:
            tree = ET.parse(file_path)
            return tree.getroot()
        except ET.ParseError as e:
            self.logger.error(f"Error parsing XML file {file_path}: {e}")
            return None
        except Exception as e:
            self.logger.error(f"Unexpected error parsing XML file {file_path}: {e}")
            return None

    async def safe_parse_json(self, file_path: Path) -> Optional[Dict[str, Any]]:
        """
        Safely parse a JSON file and return the data.

        Args:
            file_path: Path to the JSON file

        Returns:
            Parsed JSON data or None if parsing fails
        """
        try:
            with open(file_path, "r", encoding="utf-8") as f:
                return json.load(f)
        except json.JSONDecodeError as e:
            self.logger.error(f"Error parsing JSON file {file_path}: {e}")
            return None
        except Exception as e:
            self.logger.error(f"Unexpected error parsing JSON file {file_path}: {e}")
            return None

    def extract_cves_from_text(self, text: str) -> List[str]:
        """
        Extract CVE IDs from text.

        Args:
            text: Text to search for CVE IDs

        Returns:
            List of CVE IDs found
        """
        cve_pattern = r"CVE-\d{4}-\d+(?:\.\d+)?"
        return [cve.upper() for cve in re.findall(cve_pattern, text)]

    def build_description(self, title: str, details: Dict[str, str]) -> str:
        """
        Build a structured markdown description from a set of details.

        Args:
            title: Title of the vulnerability
            details: Dictionary of section name to content

        Returns:
            Formatted markdown description
        """
        description = f"## {title}\n\n"

        for section, content in details.items():
            if content and content.strip():
                description += f"### {section}\n{content}\n\n"

        return description
