import json
import xml.etree.ElementTree as ET
from pathlib import Path
from typing import Dict, Type, Any, Optional
import importlib
import inspect

from .base import BaseParser
from ..utils.logging import get_logger


class ParserFactory:
    """
    Factory class for creating parser instances based on file type detection.

    This factory automatically detects the file format and scanner type,
    then creates and returns an appropriate parser instance.
    """

    def __init__(self):
        self.logger = get_logger("ParserFactory")
        self._scanner_types = self._load_parser_classes()

        # Define file type detection rules
        self.file_type_rules = [
            (lambda p: p.suffix.lower() == ".xml", "xml"),
            (lambda p: p.suffix.lower() == ".json", "json"),
            (
                lambda p: p.suffix.lower() == ".html" or p.suffix.lower() == ".htm",
                "html",
            ),
            (lambda p: p.suffix.lower() == ".txt", "text"),
            (lambda p: p.suffix.lower() == ".csv", "csv"),
            (lambda p: p.suffix.lower() == ".gvr", "xml"),
        ]

        # Define scanner detection rules for each file format
        self.scanner_detection_rules = {
            "xml": [
                (lambda root: root.tag == "nmaprun", "nmap"),
                (lambda root: "nuclei" in str(root.attrib), "nuclei"),
                (
                    lambda root: root.tag == "report"
                    and (
                        root.find(".//results") is not None
                        or root.find(".//result") is not None
                    )
                    and root.find(".//nvt") is not None,
                    "openvas",
                ),
            ],
            "json": [
                (
                    lambda data: "site" in data
                    and ("@name" in data["site"] or "@host" in data["site"]),
                    "zap",
                ),
                (lambda data: "template" in data and "info" in data, "nuclei"),
            ],
            "html": [
                (lambda content: "OpenVAS" in content or "GVM" in content, "openvas"),
            ],
        }

    def _load_parser_classes(self) -> Dict[str, Type[BaseParser]]:
        """
        Dynamically load all parser classes from the parsers module using reflection.

        This method:
        1. Scans the current directory for Python files that might contain parsers
        2. Excludes special files like __init__.py, base.py, and factory.py
        3. Imports each module dynamically using importlib
        4. Searches each module for classes that inherit from BaseParser
        5. Registers each parser with a normalized name (removing _parser suffix)

        This approach allows new parsers to be added by simply creating a new file
        in the parsers directory without modifying existing code, following the
        Open/Closed Principle of software design.

        Returns:
            Dictionary mapping scanner types to their parser classes
        """
        scanner_types = {}

        # Get the current parsers directory
        parsers_dir = Path(__file__).parent

        # Find all Python files in the parsers directory
        parser_files = list(parsers_dir.glob("*.py"))

        for parser_file in parser_files:
            # Skip __init__, base, and factory modules
            if parser_file.stem in ["__init__", "base", "factory"]:
                continue

            try:
                # Normalize the scanner name (remove _parser suffix if present)
                scanner_name = parser_file.stem
                if scanner_name.endswith("_parser"):
                    scanner_name = scanner_name[:-7]

                # Import the module
                module_name = f".{parser_file.stem}"
                module = importlib.import_module(module_name, package="src.parsers")

                # Find the parser class in the module
                for item_name in dir(module):
                    item = getattr(module, item_name)

                    # Check if it's a class, subclass of BaseParser, and not BaseParser itself
                    if (
                        inspect.isclass(item)
                        and issubclass(item, BaseParser)
                        and item != BaseParser
                    ):

                        scanner_types[scanner_name] = item
                        self.logger.debug(
                            f"Registered parser for {scanner_name}: {item.__name__}"
                        )
                        break

            except Exception as e:
                self.logger.warning(f"Error loading parser from {parser_file}: {e}")

        return scanner_types

    async def create_parser(self, file_path: Path) -> Optional[BaseParser]:
        """
        Create a parser instance for the given file.

        Args:
            file_path: Path to the vulnerability report file

        Returns:
            An appropriate parser instance or None if file type is not supported
        """
        file_type = self._detect_file_type(file_path)
        scanner_type = await self._detect_scanner_type(file_path, file_type)

        self.logger.info(
            f"File {file_path.name} detected as {file_type} format from {scanner_type} scanner"
        )

        if scanner_type in self._scanner_types:
            return self._scanner_types[scanner_type]()
        else:
            self.logger.warning(f"No parser available for {scanner_type} scanner")
            return None

    def _detect_file_type(self, file_path: Path) -> str:
        """
        Detect the file type based on extension and content.

        Args:
            file_path: Path to the file

        Returns:
            Detected file type (xml, json, html, etc.)
        """
        # Check file extension using our rules
        for rule, file_type in self.file_type_rules:
            if rule(file_path):
                return file_type

        # If extension is not conclusive, check content
        try:
            with open(file_path, "r", encoding="utf-8") as f:
                content_start = f.read(1024)  # Read first 1KB

                if content_start.strip().startswith("{"):
                    return "json"
                elif content_start.strip().startswith("<"):
                    return "xml"
                elif "<html" in content_start.lower():
                    return "html"
                else:
                    # Default to text if we can't determine
                    return "text"
        except UnicodeDecodeError:
            # Try reading as binary for binary formats
            return "binary"

    async def _detect_scanner_type(self, file_path: Path, file_type: str) -> str:
        """
        Detect which vulnerability scanner produced this file.

        Args:
            file_path: Path to the file
            file_type: Previously detected file type

        Returns:
            Scanner type string (nmap, nuclei, etc.)
        """
        # Use file-type specific detection rules
        if file_type == "xml":
            try:
                tree = ET.parse(file_path)
                root = tree.getroot()

                # Check each rule for XML files
                for rule, scanner_type in self.scanner_detection_rules.get("xml", []):
                    if rule(root):
                        return scanner_type

            except ET.ParseError:
                self.logger.warning(f"Failed to parse {file_path} as XML")

        elif file_type == "json":
            try:
                with open(file_path, "r", encoding="utf-8") as f:
                    # Read only 8KB to avoid loading large files completely
                    data = json.loads(f.read(8192))

                    # Check each rule for JSON files
                    for rule, scanner_type in self.scanner_detection_rules.get(
                        "json", []
                    ):
                        if rule(data):
                            return scanner_type

            except (json.JSONDecodeError, UnicodeDecodeError):
                self.logger.warning(f"Failed to parse {file_path} as JSON")

        elif file_type == "html":
            try:
                with open(file_path, "r", encoding="utf-8") as f:
                    content = f.read(4096)  # Read first 4KB

                    # Check each rule for HTML files
                    for rule, scanner_type in self.scanner_detection_rules.get(
                        "html", []
                    ):
                        if rule(content):
                            return scanner_type

            except UnicodeDecodeError:
                self.logger.warning(f"Failed to parse {file_path} as HTML")

        # Fallback: try all parsers to see if any can validate the file
        for scanner_type, parser_class in self._scanner_types.items():
            parser = parser_class()
            if await parser.validate_file(file_path):
                return scanner_type

        # Default to unknown if we can't determine
        return "unknown"
