import xml.etree.ElementTree as ET
from pathlib import Path
from typing import List, Dict, Any, Optional
import re
from datetime import datetime

from .base import BaseParser
from ..models.vulnerability import (
    Vulnerability,
    Severity,
    AffectedComponent,
)


class NmapParser(BaseParser):
    """Parser for Nmap XML output with enhanced vulnerability detection."""

    def __init__(self):
        super().__init__()
        self.service_to_vulns = {}  # Mapping services -> vulnerabilities

    async def _process_service(
        self, service_elem: ET.Element, host: str, port: str
    ) -> List[Vulnerability]:
        """Process a service element to identify vulnerabilities."""
        vulnerabilities = []

        product = service_elem.get("product", "")
        version = service_elem.get("version", "")

        if product and version:
            service_id = f"{product}:{version}"
            vuln = Vulnerability(
                id=f"SERVICE-{host}-{port}-{product}-{version}",
                title=f"Detected {product} {version}",
                description=(
                    f"Detected {product} version {version} running on "
                    f"host {host} port {port}"
                ),
                severity=Severity.INFO,
                affected_components=[
                    AffectedComponent(
                        name=product, version=version, location=f"{host}:{port}"
                    )
                ],
                discovered_by=["Nmap"],
                discovery_date=self.get_current_timestamp(),
                last_seen_date=self.get_current_timestamp(),
            )
            vulnerabilities.append(vuln)
            self.service_to_vulns[service_id] = vuln

        return vulnerabilities

    async def _process_script(
        self, script_elem: ET.Element, host: str
    ) -> List[Vulnerability]:
        """Process script output, particularly focusing on vulnerability scripts."""
        vulnerabilities = []
        script_id = script_elem.get("id", "")
        output = script_elem.get("output", "")

        if script_id == "vulners":
            # Split the output into lines
            lines = [line.strip() for line in output.split("\n") if line.strip()]

            # Extract CVEs
            for line in lines:
                if "CVE-" in line:
                    try:
                        # Extract CVE and description
                        cve_match = re.search(r"(CVE-\d{4}-\d+)", line)
                        if cve_match:
                            cve_id = cve_match.group(1)
                            description = (
                                line.split(cve_id + ":", 1)[1].strip()
                                if ":" in line
                                else line
                            )

                            # Determine severity based on context
                            severity = self._determine_vulnerability_severity(
                                description, cve_id
                            )

                            # Find the affected service
                            affected_service = self._find_affected_service(description)

                            components = []
                            if affected_service:
                                components.append(
                                    AffectedComponent(
                                        name=affected_service["product"],
                                        version=affected_service["version"],
                                        location=f"{host}:{affected_service['port']}",
                                    )
                                )
                            else:
                                components.append(
                                    AffectedComponent(name=host, location=host)
                                )

                            vuln = Vulnerability(
                                id=cve_id,
                                title=f"Vulnerability {cve_id}",
                                description=description,
                                severity=severity,
                                affected_components=components,
                                discovered_by=["Nmap Vulners"],
                                discovery_date=self.get_current_timestamp(),
                                last_seen_date=self.get_current_timestamp(),
                            )
                            vulnerabilities.append(vuln)

                    except Exception as e:
                        self.logger.error(
                            f"Error processing vulnerability line: {line}"
                        )
                        self.logger.error(f"Error details: {e}")
                        continue

        return vulnerabilities

    async def parse_file(self, file_path: Path) -> List[Vulnerability]:
        """Parse an Nmap XML report file."""
        vulnerabilities = []

        try:
            tree = ET.parse(file_path)
            root = tree.getroot()

            # Reset the service mapping for each new file
            self.service_to_vulns = {}

            # Process each host
            for host in root.findall(".//host"):
                # Get host address
                address = host.find(".//address").get("addr")

                # Process services
                for port in host.findall(".//port"):
                    service = port.find("service")
                    if service is not None:
                        service_vulns = await self._process_service(
                            service, address, port.get("portid")
                        )
                        vulnerabilities.extend(service_vulns)

                # Process vulnerability scripts
                for script in host.findall(".//hostscript/script"):
                    script_vulns = await self._process_script(script, address)
                    vulnerabilities.extend(script_vulns)

            self.logger.info(
                f"Found {len(vulnerabilities)} vulnerabilities in {file_path}"
            )
            return vulnerabilities

        except ET.ParseError as e:
            self.logger.error(f"Error parsing XML file {file_path}: {e}")
            raise
        except Exception as e:
            self.logger.error(f"Unexpected error processing {file_path}: {e}")
            raise

    async def validate_file(self, file_path: Path) -> bool:
        """Validate if the file is a valid Nmap XML report."""
        if not file_path.suffix.lower() == ".xml":
            return False

        try:
            tree = ET.parse(file_path)
            root = tree.getroot()
            return root.tag == "nmaprun"
        except ET.ParseError:
            return False
        except Exception:
            return False

    def _determine_vulnerability_severity(
        self, description: str, cve_id: str
    ) -> Severity:
        """Determine vulnerability severity based on context and CVE ID."""
        description_lower = description.lower()

        # Keywords indicating critical severity
        if any(
            word in description_lower
            for word in [
                "remote code execution",
                "rce",
                "arbitrary code",
                "critical",
                "arbitrary command",
            ]
        ):
            return Severity.CRITICAL

        # Keywords indicating high severity
        if any(
            word in description_lower
            for word in [
                "privilege escalation",
                "authentication bypass",
                "sql injection",
                "arbitrary file",
                "buffer overflow",
            ]
        ):
            return Severity.HIGH

        # Keywords indicating medium severity
        if any(
            word in description_lower
            for word in [
                "cross-site scripting",
                "xss",
                "information disclosure",
                "denial of service",
                "dos",
            ]
        ):
            return Severity.MEDIUM

        # Keywords indicating low severity
        if any(word in description_lower for word in ["local", "minor", "low impact"]):
            return Severity.LOW

        return Severity.MEDIUM  # Default to MEDIUM if unsure

    def _find_affected_service(self, description: str) -> Optional[Dict[str, str]]:
        """Find the service affected by a vulnerability based on description."""
        description_lower = description.lower()

        for service_id, vuln in self.service_to_vulns.items():
            product, version = service_id.split(":")
            if product.lower() in description_lower:
                port = (
                    vuln.affected_components[0].location.split(":")[1]
                    if vuln.affected_components
                    else None
                )
                return {"product": product, "version": version, "port": port}

        return None
