import xml.etree.ElementTree as ET
import json
from pathlib import Path
from typing import List, Dict, Any, Optional
import re
import urllib.parse
from datetime import datetime

from .base import BaseParser
from ..models.vulnerability import (
    Vulnerability,
    AffectedComponent,
)


class NucleiParser(BaseParser):
    """Parser for Nuclei scanner reports (XML and JSON formats)."""

    async def parse_file(self, file_path: Path) -> List[Vulnerability]:
        """Parse a Nuclei report file."""
        self.logger.info(f"Parsing Nuclei report file: {file_path}")

        # Determine file format
        if file_path.suffix.lower() == ".xml":
            return await self._parse_xml(file_path)
        elif file_path.suffix.lower() == ".json":
            return await self._parse_json(file_path)
        elif file_path.suffix.lower() == ".txt":
            return await self._parse_txt(file_path)
        else:
            self.logger.warning(
                f"Unsupported file format for Nuclei parser: {file_path.suffix}"
            )
            return []

    # Update the validation method in NucleiParser
    async def validate_file(self, file_path: Path) -> bool:
        """
        Validate if the file is a valid Nuclei report.

        Args:
            file_path: Path to the file to validate

        Returns:
            True if file is valid, False otherwise
        """

        # Check for text format
        if file_path.suffix.lower() == ".txt":
            try:
                with open(file_path, "r") as f:
                    # Check first few lines for Nuclei patterns
                    for i, line in enumerate(f):
                        if i > 5:  # Check first 5 lines
                            break
                        # Look for typical Nuclei format: [template] [protocol] [severity]
                        if re.match(r"\[(.*?)\]\s+\[(.*?)\]\s+\[(.*?)\]", line):
                            return True
                return False
            except Exception:
                return False

        # Check XML files
        if file_path.suffix.lower() == ".xml":
            try:
                tree = ET.parse(file_path)
                root = tree.getroot()

                # Look for specific Nuclei XML patterns
                # Check for rows with nuclei template names
                for row in root.findall(".//row"):
                    for child in row:
                        if child.tag and "http" in child.tag and "CVE" in child.tag:
                            return True

                return False
            except Exception:
                return False

        # Check JSON files
        elif file_path.suffix.lower() == ".json":
            try:
                # Check first few lines for Nuclei patterns
                with open(file_path, "r") as f:
                    # Try to detect JSON format - either a JSON array or JSON lines
                    first_char = f.read(1)
                    f.seek(0)

                    if first_char == "[":
                        # Standard JSON array
                        data = json.load(f)
                        if isinstance(data, list) and len(data) > 0:
                            # Check both template and template-id keys
                            if ("template" in data[0] and "info" in data[0]) or (
                                "template-id" in data[0] and "info" in data[0]
                            ):
                                return True
                    else:
                        # JSON lines - each line is a separate JSON object
                        for i, line in enumerate(f):
                            if i > 10:  # Check first 10 lines max
                                break
                            try:
                                obj = json.loads(line.strip())
                                # Check both template and template-id keys
                                if ("template" in obj and "info" in obj) or (
                                    "template-id" in obj and "info" in obj
                                ):
                                    return True
                            except:
                                continue

            except Exception:
                return False

        return False

        return False

    async def _parse_txt(self, file_path: Path) -> List[Vulnerability]:
        """Parse a Nuclei classic text report file."""
        vulnerabilities = []
        timestamp = self.get_current_timestamp()

        try:
            with open(file_path, "r") as f:
                for line in f:
                    line = line.strip()
                    if not line:
                        continue

                    # Parse the line using regex
                    # Format is typically [id] [protocol] [severity] target [optional_data]
                    match = re.match(
                        r"\[(.*?)\]\s+\[(.*?)\]\s+\[(.*?)\]\s+(.*?)(?:\s+\[(.*)\])?$",
                        line,
                    )
                    if not match:
                        self.logger.warning(f"Could not parse Nuclei line: {line}")
                        continue

                    template_id, protocol, severity, target, extra_data = match.groups()

                    # Extract host and possibly URL
                    if "://" in target:
                        # It's a URL
                        url = target
                        parsed_url = urllib.parse.urlparse(url)
                        host = parsed_url.netloc
                        path = parsed_url.path + (
                            "?" + parsed_url.query if parsed_url.query else ""
                        )
                    else:
                        # It's a host:port
                        url = None
                        host_parts = target.split(":")
                        host = host_parts[0]
                        path = f":{host_parts[1]}" if len(host_parts) > 1 else ""

                    # Check if it's a CVE
                    cve_id = None
                    if template_id.startswith("CVE-"):
                        cve_id = template_id

                    # Parse extra data if available
                    extra_info = {}
                    if extra_data:
                        # Handle simple key-value pairs in brackets
                        try:
                            # Try to parse as JSON-like format
                            if extra_data.startswith('"') and extra_data.endswith('"'):
                                # It's a list of values
                                values = extra_data.strip("[]").split('","')
                                extra_info["values"] = [v.strip('"') for v in values]
                            elif "=" in extra_data:
                                # Key-value pairs
                                pairs = extra_data.strip("[]").split(",")
                                for pair in pairs:
                                    if "=" in pair:
                                        k, v = pair.split("=", 1)
                                        extra_info[k] = v.strip('"')
                        except:
                            # If parsing fails, just store as raw data
                            extra_info["raw"] = extra_data

                    # Build vulnerability details
                    title = f"Nuclei {template_id}"
                    description = f"## {title}\n\n"
                    description += f"**Protocol**: {protocol}\n"
                    description += f"**Severity**: {severity}\n"
                    description += f"**Target**: {target}\n\n"

                    if extra_info:
                        description += "**Additional Information**:\n"
                        for k, v in extra_info.items():
                            if isinstance(v, list):
                                description += f"- {k}: {', '.join(v)}\n"
                            else:
                                description += f"- {k}: {v}\n"

                    # Create vulnerability
                    vuln = Vulnerability(
                        id=cve_id if cve_id else f"NUCLEI-{template_id}",
                        title=title,
                        description=description,
                        severity=self.normalize_severity(severity),
                        affected_components=[
                            AffectedComponent(name=protocol.upper(), location=target)
                        ],
                        discovered_by=["Nuclei"],
                        discovery_date=timestamp,
                        last_seen_date=timestamp,
                        references=[],
                        tags=[template_id, protocol],
                        metadata={
                            "template": template_id,
                            "protocol": protocol,
                            "related_cves": [cve_id] if cve_id else [],
                        },
                    )
                    vulnerabilities.append(vuln)

            return vulnerabilities

        except Exception as e:
            self.logger.error(f"Error parsing Nuclei text file {file_path}: {e}")
            return []

    async def _parse_xml(self, file_path: Path) -> List[Vulnerability]:
        """Parse a Nuclei XML report file."""
        vulnerabilities = []

        try:
            tree = ET.parse(file_path)
            root = tree.getroot()

            timestamp = self.get_current_timestamp()

            # Process each row (finding)
            for row in root.findall(".//row"):
                for child in row:
                    if child.tag and child.text:
                        vuln = self._parse_nuclei_finding(
                            child.tag, child.text, timestamp
                        )
                        if vuln:
                            vulnerabilities.append(vuln)

            return vulnerabilities

        except ET.ParseError as e:
            self.logger.error(f"Error parsing XML file {file_path}: {e}")
            raise
        except Exception as e:
            self.logger.error(f"Unexpected error processing {file_path}: {e}")
            raise

    # Update the _parse_json method in NucleiParser
    async def _parse_json(self, file_path: Path) -> List[Vulnerability]:
        """Parse a Nuclei JSON report file."""
        vulnerabilities = []

        try:
            # Detect format - array or JSON lines
            with open(file_path, "r") as f:
                first_char = f.read(1)
                f.seek(0)

                findings = []

                if first_char == "[":
                    # Standard JSON array
                    findings = json.load(f)
                else:
                    # JSON lines - each line is a separate JSON object
                    for line in f:
                        try:
                            obj = json.loads(line.strip())
                            findings.append(obj)
                        except:
                            continue

            timestamp = self.get_current_timestamp()

            # Process each finding
            for finding in findings:
                # Get template identifier from either template or template-id field
                template = finding.get("template", finding.get("template-id", ""))
                template_id = finding.get("template-id", "")
                matched = finding.get("matched", finding.get("matched-at", "Unknown"))

                # Extract metadata
                info = finding.get("info", {})
                name = info.get("name", "Unknown Nuclei Finding")
                severity = info.get("severity", "medium")
                description = info.get("description", "")

                tags = info.get("tags", [])
                references = info.get("reference", [])
                cves = []

                # Extract CVEs from tags or references
                for tag in tags:
                    if tag.startswith("cve:"):
                        cve_id = tag.split(":", 1)[1].upper()
                        if not cve_id.startswith("CVE-"):
                            cve_id = f"CVE-{cve_id}"
                        cves.append(cve_id)

                # Check for CVE-ID in classification section if it exists
                classification = info.get("classification", {})
                if classification:
                    cve_ids = classification.get("cve-id", [])
                    if isinstance(cve_ids, list):
                        for cve_id in cve_ids:
                            if cve_id and cve_id.lower() != "null":
                                formatted_cve = cve_id.upper()
                                if not formatted_cve.startswith("CVE-"):
                                    formatted_cve = f"CVE-{formatted_cve}"
                                cves.append(formatted_cve)

                # Extract CVEs from references
                for ref in references:
                    cve_match = re.search(r"CVE-\d{4}-\d+", ref, re.IGNORECASE)
                    if cve_match:
                        cves.append(cve_match.group(0).upper())

                # Build description
                full_description = f"## {name}\n\n"

                if description:
                    full_description += f"### Description\n{description}\n\n"

                full_description += f"**Severity**: {severity.capitalize()}\n"
                full_description += f"**Template**: {template}\n"
                full_description += f"**Matched URL**: {matched}\n\n"

                if cves:
                    full_description += f"**Associated CVEs**: {', '.join(cves)}\n"

                if tags:
                    full_description += f"**Tags**: {', '.join(tags)}\n"

                if references:
                    full_description += f"### References\n"
                    for ref in references:
                        full_description += f"- {ref}\n"

                # Determine ID
                vuln_id = (
                    cves[0]
                    if cves
                    else (
                        f"NUCLEI-{template_id}" if template_id else f"NUCLEI-{template}"
                    )
                )

                # Create vulnerability
                vuln = Vulnerability(
                    id=vuln_id,
                    title=name,
                    description=full_description,
                    severity=self.normalize_severity(severity),
                    affected_components=[
                        AffectedComponent(name="Web Application", location=matched)
                    ],
                    discovered_by=["Nuclei"],
                    discovery_date=timestamp,
                    last_seen_date=timestamp,
                    references=references,
                    tags=tags,
                    metadata={
                        "template": template,
                        "template_id": template_id,
                        "related_cves": cves,
                    },
                )
                vulnerabilities.append(vuln)

        except json.JSONDecodeError as e:
            self.logger.error(f"Error parsing JSON file {file_path}: {e}")
            # Just continue instead of raising the exception
        except Exception as e:
            self.logger.error(f"Unexpected error processing {file_path}: {e}")
            # Just continue instead of raising the exception

        return vulnerabilities

    def _parse_nuclei_finding(
        self, tag: str, text: str, timestamp: datetime
    ) -> Optional[Vulnerability]:
        """
        Parse a Nuclei finding from XML.

        Args:
            tag: XML tag name
            text: XML text content
            timestamp: Timestamp to use for discovery/last seen

        Returns:
            Vulnerability object or None if parsing failed
        """
        try:
            # Extract information from the tag name
            # Tags typically look like: _CVE_2021_1234_http_high_http_example_com_index_php
            tag_parts = tag.strip("_").split("_")

            # Check if this is a CVE tag
            if len(tag_parts) < 3 or tag_parts[0] != "CVE":
                return None

            # Extract CVE ID
            cve_id = f"CVE-{tag_parts[1]}-{tag_parts[2]}"

            # Extract severity if present
            severity = None
            for sev in ["critical", "high", "medium", "low", "info"]:
                if sev in tag_parts:
                    severity = sev
                    break

            if not severity:
                severity = "medium"  # Default

            # Extract URL from text if present
            url_match = re.search(r"https?://[^\s]+", text)
            location = url_match.group(0) if url_match else "Unknown"

            # Extract template name if possible
            template_name = None
            for part in tag_parts:
                if (
                    part != "CVE"
                    and part != "http"
                    and part not in ["critical", "high", "medium", "low", "info"]
                ):
                    template_name = part
                    break

            title = (
                f"Nuclei found {cve_id}"
                if template_name is None
                else f"Nuclei found {template_name} ({cve_id})"
            )

            # Build description
            description = f"## {title}\n\n"
            description += f"**Severity**: {severity.capitalize()}\n"
            description += f"**Location**: {location}\n\n"
            description += f"**Finding**: {text}\n\n"
            description += f"**CVE**: {cve_id}\n"

            # Create vulnerability
            return Vulnerability(
                id=cve_id,
                title=title,
                description=description,
                severity=self.normalize_severity(severity),
                affected_components=[
                    AffectedComponent(name="Web Application", location=location)
                ],
                discovered_by=["Nuclei"],
                discovery_date=timestamp,
                last_seen_date=timestamp,
                tags=["nuclei"],
            )

        except Exception as e:
            self.logger.error(f"Error parsing Nuclei finding: {e}")
            return None

    def _parse_nuclei_json_finding(
        self, finding: Dict[str, Any], timestamp: datetime
    ) -> Optional[Vulnerability]:
        """
        Parse a Nuclei finding from JSON.

        Args:
            finding: JSON object for the finding
            timestamp: Timestamp to use for discovery/last seen

        Returns:
            Vulnerability object or None if parsing failed
        """
        try:
            # Extract template information
            template = finding.get("template", "")
            template_id = finding.get("template-id", "")
            matched = finding.get("matched", "Unknown")

            # Extract metadata
            info = finding.get("info", {})
            name = info.get("name", "Unknown Nuclei Finding")
            severity = info.get("severity", "medium")
            description = info.get("description", "")

            tags = info.get("tags", [])
            references = info.get("reference", [])
            cves = []

            # Extract CVEs from tags or references
            for tag in tags:
                if tag.startswith("cve:"):
                    cve_id = tag.split(":", 1)[1].upper()
                    if not cve_id.startswith("CVE-"):
                        cve_id = f"CVE-{cve_id}"
                    cves.append(cve_id)

            # Extract CVEs from references
            for ref in references:
                cve_match = re.search(r"CVE-\d{4}-\d+", ref)
                if cve_match:
                    cves.append(cve_match.group(0))

            # Build description
            full_description = f"## {name}\n\n"

            if description:
                full_description += f"### Description\n{description}\n\n"

            full_description += f"**Severity**: {severity.capitalize()}\n"
            full_description += f"**Template**: {template}\n"
            full_description += f"**Matched URL**: {matched}\n\n"

            if cves:
                full_description += f"**Associated CVEs**: {', '.join(cves)}\n"

            if tags:
                full_description += f"**Tags**: {', '.join(tags)}\n"

            if references:
                full_description += f"### References\n"
                for ref in references:
                    full_description += f"- {ref}\n"

            # Determine ID
            vuln_id = (
                cves[0]
                if cves
                else f"NUCLEI-{template_id}" if template_id else f"NUCLEI-{template}"
            )

            # Create vulnerability
            return Vulnerability(
                id=vuln_id,
                title=name,
                description=full_description,
                severity=self.normalize_severity(severity),
                affected_components=[
                    AffectedComponent(name="Web Application", location=matched)
                ],
                discovered_by=["Nuclei"],
                discovery_date=timestamp,
                last_seen_date=timestamp,
                references=references,
                tags=tags,
                metadata={
                    "template": template,
                    "template_id": template_id,
                    "related_cves": cves,
                },
            )

        except Exception as e:
            self.logger.error(f"Error parsing Nuclei JSON finding: {e}")
            return None
