import xml.etree.ElementTree as ET
from pathlib import Path
from typing import List, Optional
from datetime import datetime
import re

from .base import BaseParser
from ..models.vulnerability import (
    Vulnerability,
    Severity,
    AffectedComponent,
    CVSSMetrics,
)


class OpenVASParser(BaseParser):
    """Parser for OpenVAS/GVM XML reports."""

    async def parse_file(self, file_path: Path) -> List[Vulnerability]:
        """
        Parse an OpenVAS XML report file.

        Args:
            file_path: Path to the OpenVAS report file

        Returns:
            List of parsed Vulnerability objects
        """
        self.logger.info(f"Parsing OpenVAS report file: {file_path}")

        vulnerabilities = []

        try:
            tree = ET.parse(file_path)
            root = tree.getroot()

            timestamp = self.get_current_timestamp()

            # Process each result (vulnerability finding)
            for result in root.findall(".//result"):
                vuln = self._parse_result(result, timestamp)
                if vuln:
                    vulnerabilities.append(vuln)

            self.logger.info(
                f"Found {len(vulnerabilities)} vulnerabilities in {file_path}"
            )
            return vulnerabilities

        except ET.ParseError as e:
            self.logger.error(f"Error parsing XML file {file_path}: {e}")
            raise
        except Exception as e:
            self.logger.error(f"Unexpected error processing {file_path}: {e}")
            raise

    async def validate_file(self, file_path: Path) -> bool:
        """
        Validate if the file is a valid OpenVAS report.

        Args:
            file_path: Path to the file to validate

        Returns:
            True if file is valid, False otherwise
        """
        if not file_path.suffix.lower() in [".xml", ".gvr"]:
            return False

        try:
            tree = ET.parse(file_path)
            root = tree.getroot()

            # Check for OpenVAS report format
            # OpenVAS/GVM reports have specific elements
            return root.tag == "report" and (
                root.find(".//results") is not None
                or root.find(".//result") is not None
            )
        except Exception:
            return False

    def _parse_result(
        self, result: ET.Element, timestamp: datetime
    ) -> Optional[Vulnerability]:
        """
        Parse a single OpenVAS result element.

        Args:
            result: XML element for the result
            timestamp: Timestamp to use for discovery/last seen

        Returns:
            Vulnerability object or None if parsing failed
        """
        try:
            # Extract basic information
            name = self._get_element_text(result, "name", "Unknown OpenVAS Finding")
            host_elem = result.find("./host")

            # Use a default host if not found
            if host_elem is None:
                self.logger.debug(f"Missing host information for {name}, using default")
                # Default host when not specified
                host = "unknown_host"

                # Try to find port information at least
                port = self._get_element_text(result, "port", "unknown_port")

                # Check if there's an IP or hostname in the result somewhere
                for elem in result.iter():
                    if elem.tag in ["ip", "hostname", "target", "host_ip", "host_name"]:
                        if elem.text and elem.text.strip():
                            host = elem.text.strip()
                            break
            else:
                host = host_elem.text
                port = self._get_element_text(result, "port", "")

            # Get threat level and convert to severity
            threat = self._get_element_text(result, "threat", "Medium")
            severity_score = self._get_element_text(result, "severity", "5.0")
            severity = self._get_severity_from_threat(threat)

            # Extract NVT (Network Vulnerability Test) information
            nvt_elem = result.find("./nvt")
            if nvt_elem is None:
                self.logger.debug(
                    f"Missing NVT information for {name}, using available data"
                )
                # Create a basic vulnerability with the information we have
                oid = "unknown"
                family = "General"
                refs = []
                cve_ids = []
                description = f"OpenVAS found a vulnerability but didn't provide detailed information. Host: {host}, Port: {port}"

                # Try to extract CVEs from the result element or other places
                result_text = ET.tostring(
                    result, encoding="utf-8", method="text"
                ).decode("utf-8")
                cve_matches = re.findall(r"CVE-\d{4}-\d+", result_text)
                if cve_matches:
                    cve_ids = list(set(cve_matches))

                # Try to find a severity in the result
                severity_elem = result.find("./severity")
                severity_score = "5.0"  # Default medium
                if severity_elem is not None and severity_elem.text:
                    severity_score = severity_elem.text

                # Get threat level if available
                threat_elem = result.find("./threat")
                threat = "Medium"  # Default
                if threat_elem is not None and threat_elem.text:
                    threat = threat_elem.text
            else:
                oid = nvt_elem.get("oid", "")
                family = self._get_element_text(nvt_elem, "family", "")

                # Extract references
                refs = []
                refs_elem = nvt_elem.find("./refs")
                if refs_elem is not None:
                    for ref_elem in refs_elem.findall("./ref"):
                        ref_id = ref_elem.get("id", "")
                        if ref_id:
                            refs.append(ref_id)

                # Extract CVE IDs from references
                cve_ids = []
                for ref in refs:
                    if ref.startswith("CVE-"):
                        cve_ids.append(ref)

                # Extract tags for additional information
                tags_text = ""
                tags_elem = nvt_elem.find("./tags")
                if tags_elem is not None and tags_elem.text:
                    tags_text = tags_elem.text

                # Parse tags into dictionary
                tag_data = {}
                for tag_pair in tags_text.split("|"):
                    if "=" in tag_pair:
                        key, value = tag_pair.split("=", 1)
                        tag_data[key] = value

                # Extract description from tags
                description = tag_data.get("summary", "")
                impact = tag_data.get("impact", "")
                solution = tag_data.get("solution", "")
                affected = tag_data.get("affected", "")
                insight = tag_data.get("insight", "")

            # Build comprehensive description
            full_description = f"## {name}\n\n"

            if description:
                full_description += f"### Description\n{description}\n\n"

            if "affected" in locals() and affected:
                full_description += f"### Affected\n{affected}\n\n"

            if "impact" in locals() and impact:
                full_description += f"### Impact\n{impact}\n\n"

            if "insight" in locals() and insight:
                full_description += f"### Insight\n{insight}\n\n"

            if "solution" in locals() and solution:
                full_description += f"### Solution\n{solution}\n\n"

            # Build affected component
            location = f"{host}:{port}" if port else host
            affected_components = [
                AffectedComponent(
                    name=f"{family} service" if family else "System", location=location
                )
            ]

            # Extract CVSS information
            cvss = None
            if nvt_elem is not None:
                cvss_base = nvt_elem.find("./cvss_base")
                if cvss_base is not None and cvss_base.text:
                    base_score = float(cvss_base.text)

                    # Look for vector string in severities
                    vector_string = ""
                    severities = nvt_elem.find("./severities")
                    if severities is not None:
                        for severity_elem in severities.findall("./severity"):
                            if severity_elem.get("type") == "cvss_base_v2":
                                value_elem = severity_elem.find("./value")
                                if value_elem is not None and value_elem.text:
                                    vector_string = value_elem.text
                                    break

                    cvss = CVSSMetrics(
                        version="2.0",
                        base_score=base_score,
                        vector_string=vector_string,
                    )
            else:
                # Try to extract CVSS from result directly
                severity_val = self._get_element_text(result, "severity", None)
                if severity_val and severity_val.replace(".", "", 1).isdigit():
                    cvss = CVSSMetrics(
                        version="2.0", base_score=float(severity_val), vector_string=""
                    )

            # Create tags list
            tags = []
            if family:
                tags.append(family.lower().replace(" ", "_"))

            # Use CVE ID as primary ID if available, otherwise use OpenVAS OID
            vuln_id = cve_ids[0] if cve_ids else f"OPENVAS-{oid.replace('.', '-')}"

            # Create the vulnerability object
            return Vulnerability(
                id=vuln_id,
                title=name,
                description=full_description,
                severity=(
                    self._get_severity_from_threat(threat)
                    if "threat" in locals()
                    else Severity.MEDIUM
                ),
                cvss=cvss,
                affected_components=[
                    AffectedComponent(
                        name=f"{family} service" if family else "System",
                        location=location,
                    )
                ],
                discovered_by=["OpenVAS"],
                discovery_date=timestamp,
                last_seen_date=timestamp,
                references=refs,
                tags=tags + cve_ids,
                metadata={
                    "oid": oid,
                    "threat": threat if "threat" in locals() else "Unknown",
                    "qod": self._get_element_text(result, "qod/value", "0"),
                    "scan_nvt_version": self._get_element_text(
                        result, "scan_nvt_version", ""
                    ),
                    "related_cves": cve_ids,
                },
            )

        except Exception as e:
            self.logger.error(f"Error parsing OpenVAS result: {e}")
            return None

    def _get_element_text(
        self, element: ET.Element, path: str, default: str = ""
    ) -> str:
        """Get text from a child element at path, or default if not found."""
        child = element.find(path)
        return child.text if child is not None and child.text else default

    def _get_severity_from_threat(self, threat: str) -> Severity:
        """Convert OpenVAS threat level to Severity enum."""
        threat = threat.lower()
        if threat == "high":
            return Severity.HIGH
        elif threat == "medium":
            return Severity.MEDIUM
        elif threat == "low":
            return Severity.LOW
        elif threat in ["log", "debug"]:
            return Severity.INFO
        elif threat == "critical":
            return Severity.CRITICAL
        else:
            return Severity.MEDIUM  # Default
