#!/usr/bin/env python3
"""
Vulnerability Processing Core

This module contains the core functionality for processing vulnerability reports.
It handles report parsing, deduplication, enrichment, and risk scoring, producing
a consolidated output file with the results.

The module provides the main processing pipeline as well as utilities for saving
the processed vulnerability data.
"""

from pathlib import Path
from typing import List, Dict, Any, Optional
import json
from datetime import datetime

from .utils.file_tracker import FileTracker
from .parsers.factory import ParserFactory
from .analysis.embedding import VulnerabilityEmbedder
from .analysis.deduplication import VulnerabilityDeduplicator
from .analysis.correlation import VulnerabilityCorrelator
from .models.vulnerability import Vulnerability
from .utils.logging import setup_logging
from .utils.exception_handler import async_exception_handler, exception_boundary


class DateTimeEncoder(json.JSONEncoder):
    """
    Custom JSON encoder for datetime objects.

    This encoder converts datetime objects to ISO format strings
    for proper JSON serialization.
    """

    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        return super().default(obj)


def save_results(vulnerabilities: List[Vulnerability], output_file: Path) -> None:
    """
    Save processed vulnerabilities to a JSON file.

    Args:
        vulnerabilities: List of vulnerabilities to save
        output_file: Output file path
    """
    # Create output directory if it doesn't exist
    output_file.parent.mkdir(parents=True, exist_ok=True)

    # Convert vulnerabilities to dict
    vuln_dicts = []
    for vuln in vulnerabilities:
        # Convert to dict and handle datetime fields explicitly
        vuln_dict = vuln.model_dump(
            exclude_none=True
        )  # Exclude fields with None values
        # Ensure all datetime fields are converted to ISO format strings
        if "discovery_date" in vuln_dict:
            vuln_dict["discovery_date"] = vuln_dict["discovery_date"].isoformat()
        if "last_seen_date" in vuln_dict:
            vuln_dict["last_seen_date"] = vuln_dict["last_seen_date"].isoformat()
        vuln_dicts.append(vuln_dict)

    # Save to file
    with open(output_file, "w") as f:
        json.dump(
            {
                "generated_at": datetime.now().isoformat(),
                "vulnerability_count": len(vulnerabilities),
                "vulnerabilities": vuln_dicts,
            },
            f,
            indent=2,
            cls=DateTimeEncoder,  # Use custom encoder for any remaining datetime objects
        )

    # Track the latest file for the HTML reporter
    FileTracker.set_latest_file(output_file)


async def process_vulnerability_reports(
    input_dir: Path, output_dir: Path, config: Dict[str, Any] = None
) -> Optional[Path]:
    # Set up default config
    if config is None:
        config = {}

    # Setup logging
    logger = setup_logging(output_dir / f"process_{datetime.now():%Y%m%d_%H%M%S}.log")
    logger.info("Starting vulnerability report processing")

    try:
        # Initialize components
        parser_factory = ParserFactory()
        deduplicator = VulnerabilityDeduplicator(
            similarity_threshold=config.get("similarity_threshold", 0.85)
        )
        correlator = VulnerabilityCorrelator()

        # Process all vulnerability report files
        all_vulnerabilities = []

        # Get all files in input directory (including subdirectories)
        input_files = list(input_dir.glob("**/*"))
        logger.info(f"Found {len(input_files)} files in input directory")

        # Filter out directories
        input_files = [f for f in input_files if f.is_file()]

        for file_path in input_files:
            logger.info(f"Processing file: {file_path}")

            try:
                # Create appropriate parser for this file
                parser = await parser_factory.create_parser(file_path)

                if parser is None:
                    logger.warning(f"No suitable parser found for {file_path}")
                    continue

                # Validate file
                if not await parser.validate_file(file_path):
                    logger.warning(f"Invalid file format: {file_path}")
                    continue

                # Parse vulnerabilities
                try:
                    vulnerabilities = await parser.parse_file(file_path)
                    all_vulnerabilities.extend(vulnerabilities)
                    logger.info(
                        f"Found {len(vulnerabilities)} vulnerabilities in {file_path}"
                    )
                except Exception as e:
                    logger.error(f"Error parsing {file_path}: {e}")
                    continue

            except Exception as e:
                logger.error(f"Error processing {file_path}: {e}")
                continue

        if not all_vulnerabilities:
            logger.warning("No vulnerabilities found in any input files")
            return None

        # Deduplicate vulnerabilities
        logger.info(
            f"Starting deduplication of {len(all_vulnerabilities)} vulnerabilities"
        )
        unique_vulnerabilities = await deduplicator.deduplicate(all_vulnerabilities)
        logger.info(
            f"After deduplication: {len(unique_vulnerabilities)} unique vulnerabilities"
        )

        # Correlate findings from different scanners
        logger.info("Correlating findings from different scanners")
        correlated_vulnerabilities = await correlator.correlate_findings(
            unique_vulnerabilities
        )
        logger.info("Correlation completed")

        # Save results
        output_file = (
            output_dir / f"vulnerabilities_{datetime.now():%Y%m%d_%H%M%S}.json"
        )
        save_results(correlated_vulnerabilities, output_file)
        logger.info(f"Results saved to {output_file}")
        return output_file

    except Exception as e:
        logger.error(f"Error during processing: {e}", exc_info=True)
        raise
