import functools
import logging
import asyncio
from typing import Callable, TypeVar

T = TypeVar("T")


def async_exception_handler(logger=None):
    """
    Decorator to handle exceptions in async functions, ensuring they are properly logged.

    This decorator wraps async functions to catch any unhandled exceptions, log them
    with detailed information about the function call, and optionally re-raise the
    exception. This is particularly useful for asyncio-based applications where
    exceptions in tasks might otherwise be lost.

    Example usage:
        @async_exception_handler(logger=my_logger)
        async def fetch_data_from_api(url):
            response = await aiohttp.ClientSession().get(url)
            return await response.json()

    If fetch_data_from_api raises an exception, it will be logged with the URL
    parameter value and full traceback before being re-raised.

    Args:
        logger: Optional logger to use for error messages. If not provided,
                a logger is created using the function's module name.

    Returns:
        Decorated function that catches and logs exceptions
    """

    def decorator(func: Callable[..., T]) -> Callable[..., T]:
        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            nonlocal logger
            if logger is None:
                logger = logging.getLogger(func.__module__)

            try:
                return await func(*args, **kwargs)
            except Exception as e:
                # Get function signature for better error reporting
                func_str = f"{func.__name__}({', '.join(repr(a) for a in args)})"
                if kwargs:
                    func_str += (
                        f", {', '.join(f'{k}={v!r}' for k, v in kwargs.items())}"
                    )

                # Log the error with traceback
                logger.error(
                    f"Exception in {func_str}: {e.__class__.__name__}: {e}",
                    exc_info=True,
                )

                # Re-raise the exception if it should not be swallowed
                raise

        return wrapper

    return decorator


def exception_boundary(reraise=False, log_level=logging.ERROR):
    """
    Create a boundary for exceptions, preventing them from propagating upward.

    Args:
        reraise: Whether to re-raise the exception after logging
        log_level: Logging level to use for error messages

    Returns:
        Decorated function that catches and optionally swallows exceptions
    """

    def decorator(func):
        @functools.wraps(func)
        async def async_wrapper(*args, **kwargs):
            logger = logging.getLogger(func.__module__)
            try:
                return await func(*args, **kwargs)
            except Exception as e:
                logger.log(
                    log_level,
                    f"Exception caught at boundary in {func.__name__}: {e}",
                    exc_info=True,
                )
                if reraise:
                    raise
                return None

        @functools.wraps(func)
        def sync_wrapper(*args, **kwargs):
            logger = logging.getLogger(func.__module__)
            try:
                return func(*args, **kwargs)
            except Exception as e:
                logger.log(
                    log_level,
                    f"Exception caught at boundary in {func.__name__}: {e}",
                    exc_info=True,
                )
                if reraise:
                    raise
                return None

        if asyncio.iscoroutinefunction(func):
            return async_wrapper
        else:
            return sync_wrapper

    return decorator
