import logging
import sys
from pathlib import Path
from datetime import datetime
from typing import Optional, Dict, Any
import json

# Definition of ANSI colors for console logging
COLORS = {
    "RESET": "\033[0m",
    "DEBUG": "\033[36m",  # Cyan
    "INFO": "\033[32m",  # Green
    "WARNING": "\033[33m",  # Yellow
    "ERROR": "\033[31m",  # Red
    "CRITICAL": "\033[41m",  # Red background
}


class ColoredFormatter(logging.Formatter):
    """Custom formatter that adds colors to console logs."""

    def format(self, record):
        # Check if we're in a terminal that supports colors
        if sys.stdout.isatty():
            levelname = record.levelname
            message = super().format(record)
            if levelname in COLORS:
                return f"{COLORS[levelname]}{message}{COLORS['RESET']}"
        return super().format(record)


def setup_logging(
    log_file: Optional[Path] = None,
    level: int = logging.INFO,
    log_to_console: bool = True,
    log_format: str = "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    detailed_format: str = "%(asctime)s - %(name)s - %(levelname)s - %(filename)s:%(lineno)d - %(message)s",
    rotate_logs: bool = True,
    max_log_files: int = 5,
) -> logging.Logger:
    """
    Setup logging configuration.

    Args:
        log_file: Optional path to log file
        level: Logging level
        log_to_console: Whether to log to console
        log_format: Basic format for console logs
        detailed_format: Detailed format for file logs
        rotate_logs: Whether to rotate logs (keeps max_log_files)
        max_log_files: Maximum number of log files to keep

    Returns:
        Configured logger instance
    """
    # Create logger
    logger = logging.getLogger("vulnerability_aggregator")
    logger.setLevel(level)

    # Clear any existing handlers
    if logger.handlers:
        logger.handlers.clear()

    # Create formatters
    console_formatter = ColoredFormatter(log_format)
    file_formatter = logging.Formatter(detailed_format)

    # Console handler
    if log_to_console:
        console_handler = logging.StreamHandler(sys.stdout)
        console_handler.setFormatter(console_formatter)
        logger.addHandler(console_handler)

    # File handler if log file specified
    if log_file:
        # Ensure log directory exists
        log_file.parent.mkdir(parents=True, exist_ok=True)

        # If rotation is enabled, archive older log files
        if rotate_logs:
            _rotate_log_files(log_file, max_log_files)

        file_handler = logging.FileHandler(log_file)
        file_handler.setFormatter(file_formatter)
        logger.addHandler(file_handler)

    # Set propagate to False to avoid duplicate logs
    logger.propagate = False

    return logger


def get_logger(name: str) -> logging.Logger:
    """
    Get a logger instance with the given name.

    Args:
        name: Logger name

    Returns:
        Logger instance
    """
    return logging.getLogger(f"vulnerability_aggregator.{name}")


def _rotate_log_files(log_file: Path, max_files: int = 5):
    """Rotate log files, keeping only the most recent ones."""
    log_dir = log_file.parent
    base_name = log_file.stem
    extension = log_file.suffix

    # Get all existing log files with the same base name
    log_files = list(log_dir.glob(f"{base_name}*{extension}"))

    # If the log file already exists, rename it with a timestamp
    if log_file in log_files:
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        archived_name = log_dir / f"{base_name}_{timestamp}{extension}"
        try:
            log_file.rename(archived_name)
        except Exception:
            # If renaming fails, just proceed with the existing file
            pass

    # Remove oldest log files if we have too many
    log_files = list(log_dir.glob(f"{base_name}*{extension}"))
    if len(log_files) >= max_files:
        # Sort by modification time (oldest first)
        log_files.sort(key=lambda x: x.stat().st_mtime)

        # Remove oldest files to stay under the limit
        for old_file in log_files[: (len(log_files) - max_files + 1)]:
            try:
                old_file.unlink()
            except Exception:
                pass


class StructuredLoggerAdapter(logging.LoggerAdapter):
    """Logger adapter that adds structured context to log messages."""

    def __init__(self, logger, extra=None):
        super().__init__(logger, extra or {})

    def process(self, msg, kwargs):
        # Always include existing extra info
        context = self.extra.copy()

        # Add any context from kwargs
        extra = kwargs.pop("extra", {})
        context.update(extra)

        # Format context as JSON if not empty
        if context:
            try:
                context_str = json.dumps(context)
                msg = f"{msg} - Context: {context_str}"
            except (TypeError, ValueError):
                # If JSON serialization fails, just append as string
                msg = f"{msg} - Context: {str(context)}"

        return msg, kwargs

    def debug(self, msg, *args, **kwargs):
        self.log(logging.DEBUG, msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        self.log(logging.INFO, msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        self.log(logging.WARNING, msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        self.log(logging.ERROR, msg, *args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        self.log(logging.CRITICAL, msg, *args, **kwargs)

    def exception(self, msg, *args, exc_info=True, **kwargs):
        self.log(logging.ERROR, msg, *args, exc_info=exc_info, **kwargs)


def get_structured_logger(
    name: str, context: Dict[str, Any] = None
) -> StructuredLoggerAdapter:
    """Get a structured logger with context information."""
    logger = get_logger(name)
    return StructuredLoggerAdapter(logger, context)
