import argparse
import os
import ssl
from pathlib import Path
from sentence_transformers import SentenceTransformer
import logging


def setup_basic_logging():
    """Set up basic logging configuration."""
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    )
    return logging.getLogger("ModelDownloader")


def download_model(model_name: str, output_dir: str, disable_ssl_verify: bool = False):
    """
    Download a sentence-transformer model and save it locally.

    Args:
        model_name: Name of the model to download
        output_dir: Directory to save the model
        disable_ssl_verify: Whether to disable SSL verification
    """
    logger = setup_basic_logging()
    logger.info(f"Downloading model: {model_name}")

    output_path = Path(output_dir) / model_name.split("/")[-1]
    os.makedirs(output_path, exist_ok=True)

    old_https_context = None

    try:
        if disable_ssl_verify:
            logger.info("SSL verification disabled")
            old_https_context = ssl._create_default_https_context
            ssl._create_default_https_context = ssl._create_unverified_context

        model = SentenceTransformer(model_name)

        logger.info(f"Saving model to: {output_path}")
        model.save(str(output_path))

        logger.info("Model downloaded and saved successfully")
        return True
    except Exception as e:
        logger.error(f"Error downloading model: {e}")
        return False
    finally:
        if disable_ssl_verify and old_https_context:
            ssl._create_default_https_context = old_https_context


def parse_args():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(
        description="Download and save a sentence-transformer model locally."
    )
    parser.add_argument(
        "--model",
        "-m",
        type=str,
        default="all-MiniLM-L6-v2",
        help="Model name or path (default: all-MiniLM-L6-v2)",
    )
    parser.add_argument(
        "--output",
        "-o",
        type=str,
        default="models",
        help="Output directory (default: models)",
    )
    parser.add_argument(
        "--disable-ssl-verify",
        action="store_true",
        help="Disable SSL certificate verification",
    )

    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()
    download_model(args.model, args.output, args.disable_ssl_verify)
